"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __exportStar = (this && this.__exportStar) || function(m, exports) {
    for (var p in m) if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p)) __createBinding(exports, m, p);
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SplTokenBonding = exports.TimeCurveConfig = exports.ExponentialCurveConfig = exports.toU128 = void 0;
const anchor = __importStar(require("@project-serum/anchor"));
const common_1 = require("@project-serum/common");
const spl_token_1 = require("@solana/spl-token");
const web3_js_1 = require("@solana/web3.js");
const spl_utils_1 = require("@strata-foundation/spl-utils");
const bn_js_1 = __importDefault(require("bn.js"));
const curves_1 = require("./curves");
__exportStar(require("./curves"), exports);
__exportStar(require("./generated/spl-token-bonding"), exports);
/**
 * Convert a number to a string avoiding scientific notation
 * @param n
 * @returns
 */
function toFixedSpecial(num, n) {
    var str = num.toFixed(n);
    if (str.indexOf("e+") === -1)
        return str;
    // if number is in scientific notation, pick (b)ase and (p)ower
    str = str
        .replace(".", "")
        .split("e+")
        .reduce(function (b, p) {
        // @ts-ignore
        return b + Array(p - b.length + 2).join(0);
    });
    if (n > 0) {
        // @ts-ignore
        str += "." + Array(n + 1).join(0);
    }
    return str;
}
/**
 * Convert a number to a 12 decimal fixed precision u128
 *
 * @param num Number to convert to a 12 decimal fixed precision BN
 * @returns
 */
function toU128(num) {
    if (bn_js_1.default.isBN(num)) {
        return num;
    }
    return new bn_js_1.default(toFixedSpecial(num, 12).replace(".", ""));
}
exports.toU128 = toU128;
/**
 * Curve configuration for c(S^(pow/frac)) + b
 */
class ExponentialCurveConfig {
    constructor({ c = 1, b = 0, pow = 1, frac = 1, }) {
        this.c = toU128(c);
        this.b = toU128(b);
        this.pow = pow;
        this.frac = frac;
        if (this.b.gt(new bn_js_1.default(0)) && this.c.gt(new bn_js_1.default(0))) {
            throw new Error("Unsupported: Cannot define an exponential function with `b`, the math to go from base to target amount becomes too hard.");
        }
    }
    toRawPrimitiveConfig() {
        return {
            exponentialCurveV0: {
                // @ts-ignore
                c: this.c,
                // @ts-ignore
                b: this.b,
                // @ts-ignore
                pow: this.pow,
                // @ts-ignore
                frac: this.frac,
            },
        };
    }
    toRawConfig() {
        return {
            definition: {
                timeV0: {
                    curves: [
                        {
                            // @ts-ignore
                            offset: new bn_js_1.default(0),
                            // @ts-ignore
                            curve: this.toRawPrimitiveConfig(),
                        },
                    ],
                },
            },
        };
    }
}
exports.ExponentialCurveConfig = ExponentialCurveConfig;
/**
 * Curve configuration that allows the curve to change parameters at discrete time offsets from the go live date
 */
class TimeCurveConfig {
    constructor() {
        this.curves = [];
    }
    addCurve(timeOffset, curve) {
        if (this.curves.length == 0 && timeOffset != 0) {
            throw new Error("First time offset must be 0");
        }
        this.curves.push({
            curve,
            offset: new bn_js_1.default(timeOffset),
        });
        return this;
    }
    toRawConfig() {
        return {
            definition: {
                timeV0: {
                    // @ts-ignore
                    curves: this.curves.map(({ curve, offset }) => ({
                        curve: curve.toRawPrimitiveConfig(),
                        offset,
                    })),
                },
            },
        };
    }
}
exports.TimeCurveConfig = TimeCurveConfig;
function toNumber(numberOrBn, mint) {
    if (bn_js_1.default.isBN(numberOrBn)) {
        return (0, curves_1.amountAsNum)(numberOrBn, mint);
    }
    else {
        return numberOrBn;
    }
}
function toBN(numberOrBn, mint) {
    if (bn_js_1.default.isBN(numberOrBn)) {
        return numberOrBn;
    }
    else {
        return new bn_js_1.default(Math.ceil(Number(numberOrBn) * Math.pow(10, mint.decimals)));
    }
}
class SplTokenBonding extends spl_utils_1.AnchorSdk {
    constructor(provider, program) {
        super({ provider, program });
        this.curveDecoder = (pubkey, account) => {
            const coded = this.program.coder.accounts.decode("CurveV0", account.data);
            return Object.assign(Object.assign({}, coded), { publicKey: pubkey });
        };
        this.tokenBondingDecoder = (pubkey, account) => {
            const coded = this.program.coder.accounts.decode("TokenBondingV0", account.data);
            return Object.assign(Object.assign({}, coded), { publicKey: pubkey });
        };
    }
    static init(provider, splTokenBondingProgramId = SplTokenBonding.ID) {
        return __awaiter(this, void 0, void 0, function* () {
            const SplTokenBondingIDLJson = yield anchor.Program.fetchIdl(splTokenBondingProgramId, provider);
            const splTokenBonding = new anchor.Program(SplTokenBondingIDLJson, splTokenBondingProgramId, provider);
            return new this(provider, splTokenBonding);
        });
    }
    /**
     * This is an admin function run once to initialize the smart contract.
     *
     * @returns Instructions needed to create sol storage
     */
    initializeSolStorageInstructions() {
        return __awaiter(this, void 0, void 0, function* () {
            const exists = yield this.getState();
            if (exists) {
                return {
                    output: null,
                    instructions: [],
                    signers: [],
                };
            }
            const [state, bumpSeed] = yield web3_js_1.PublicKey.findProgramAddress([Buffer.from("state", "utf-8")], this.programId);
            const [solStorage, solStorageBumpSeed] = yield web3_js_1.PublicKey.findProgramAddress([Buffer.from("sol-storage", "utf-8")], this.programId);
            const [wrappedSolAuthority, mintAuthorityBumpSeed] = yield web3_js_1.PublicKey.findProgramAddress([Buffer.from("wrapped-sol-authority", "utf-8")], this.programId);
            const instructions = [];
            const signers = [];
            const mintKeypair = anchor.web3.Keypair.generate();
            signers.push(mintKeypair);
            instructions.push(...[
                web3_js_1.SystemProgram.createAccount({
                    fromPubkey: this.wallet.publicKey,
                    newAccountPubkey: mintKeypair.publicKey,
                    space: 82,
                    lamports: yield this.provider.connection.getMinimumBalanceForRentExemption(82),
                    programId: spl_token_1.TOKEN_PROGRAM_ID,
                }),
                spl_token_1.Token.createInitMintInstruction(spl_token_1.TOKEN_PROGRAM_ID, mintKeypair.publicKey, 9, this.wallet.publicKey, wrappedSolAuthority),
            ]);
            yield (0, spl_utils_1.createMetadata)(new spl_utils_1.Data({
                name: "Token Bonding Wrapped SOL",
                symbol: "twSOL",
                uri: "",
                sellerFeeBasisPoints: 0,
                // @ts-ignore
                creators: null,
            }), this.wallet.publicKey.toBase58(), mintKeypair.publicKey.toBase58(), this.wallet.publicKey.toBase58(), instructions, this.wallet.publicKey.toBase58());
            instructions.push(spl_token_1.Token.createSetAuthorityInstruction(spl_token_1.TOKEN_PROGRAM_ID, mintKeypair.publicKey, wrappedSolAuthority, "MintTokens", this.wallet.publicKey, []));
            instructions.push(yield this.instruction.initializeSolStorageV0({
                solStorageBumpSeed,
                bumpSeed,
                mintAuthorityBumpSeed,
            }, {
                accounts: {
                    state,
                    payer: this.wallet.publicKey,
                    solStorage,
                    mintAuthority: wrappedSolAuthority,
                    wrappedSolMint: mintKeypair.publicKey,
                    tokenProgram: spl_token_1.TOKEN_PROGRAM_ID,
                    systemProgram: web3_js_1.SystemProgram.programId,
                    rent: web3_js_1.SYSVAR_RENT_PUBKEY,
                },
            }));
            return {
                instructions,
                signers,
                output: null,
            };
        });
    }
    /**
     * Admin command run once to initialize the smart contract
     */
    initializeSolStorage() {
        return this.execute(this.initializeSolStorageInstructions());
    }
    /**
     * Create a curve shape for use in a TokenBonding instance
     *
     * @param param0
     * @returns
     */
    initializeCurveInstructions({ payer = this.wallet.publicKey, config: curveConfig, }) {
        return __awaiter(this, void 0, void 0, function* () {
            const curve = curveConfig.toRawConfig();
            const curveKeypair = anchor.web3.Keypair.generate();
            return {
                output: {
                    curve: curveKeypair.publicKey,
                },
                signers: [curveKeypair],
                instructions: [
                    web3_js_1.SystemProgram.createAccount({
                        fromPubkey: payer,
                        newAccountPubkey: curveKeypair.publicKey,
                        space: 500,
                        lamports: yield this.provider.connection.getMinimumBalanceForRentExemption(500),
                        programId: this.programId,
                    }),
                    yield this.instruction.createCurveV0(curve, {
                        accounts: {
                            payer,
                            curve: curveKeypair.publicKey,
                            systemProgram: web3_js_1.SystemProgram.programId,
                            rent: web3_js_1.SYSVAR_RENT_PUBKEY,
                        },
                    }),
                ],
            };
        });
    }
    /**
     * See {@link initializeCurve}
     * @param args
     * @returns
     */
    initializeCurve(args) {
        return __awaiter(this, void 0, void 0, function* () {
            return (yield this.execute(this.initializeCurveInstructions(args))).curve;
        });
    }
    /**
     * Get the PDA key of a TokenBonding given the target mint and index
     *
     * `index` = 0 is the default bonding curve that can mint `targetMint`. All other curves are curves that allow burning of `targetMint` for some different base.
     *
     * @param targetMint
     * @param index
     * @returns
     */
    tokenBondingKey(targetMint, index) {
        return __awaiter(this, void 0, void 0, function* () {
            const pad = Buffer.alloc(2);
            new bn_js_1.default(index, 16, "le").toBuffer().copy(pad);
            return web3_js_1.PublicKey.findProgramAddress([Buffer.from("token-bonding", "utf-8"), targetMint.toBuffer(), pad], this.programId);
        });
    }
    /**
     * Get the PDA key of the account that should be the authority on the base storage (reserve) account of a bonding curve that doesn't have sell frozen.
     * @param tokenBonding
     * @returns
     */
    baseStorageAuthorityKey(tokenBonding) {
        return __awaiter(this, void 0, void 0, function* () {
            return web3_js_1.PublicKey.findProgramAddress([Buffer.from("storage-authority", "utf-8"), tokenBonding.toBuffer()], this.programId);
        });
    }
    /**
     * Create a bonding curve
     *
     * @param param0
     * @returns
     */
    createTokenBondingInstructions({ generalAuthority = this.wallet.publicKey, curveAuthority = null, reserveAuthority = null, payer = this.wallet.publicKey, curve, baseMint, targetMint, baseStorage, buyBaseRoyalties, buyBaseRoyaltiesOwner = this.wallet.publicKey, buyTargetRoyalties, buyTargetRoyaltiesOwner = this.wallet.publicKey, sellBaseRoyalties, sellBaseRoyaltiesOwner = this.wallet.publicKey, sellTargetRoyalties, sellTargetRoyaltiesOwner = this.wallet.publicKey, buyBaseRoyaltyPercentage, buyTargetRoyaltyPercentage, sellBaseRoyaltyPercentage, sellTargetRoyaltyPercentage, mintCap, purchaseCap, goLiveDate = new Date(new Date().valueOf() - 1000), // 1 secs ago
    freezeBuyDate, targetMintDecimals, targetMintKeypair = web3_js_1.Keypair.generate(), buyFrozen = false, index, }) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!targetMint) {
                if (sellTargetRoyalties || buyTargetRoyalties) {
                    throw new Error("Cannot define target royalties if mint is not defined");
                }
                if (!targetMintDecimals) {
                    throw new Error("Cannot define mint without decimals ");
                }
            }
            const provider = this.provider;
            const state = (yield this.getState());
            if (baseMint.equals(spl_token_1.NATIVE_MINT)) {
                baseMint = state.wrappedSolMint;
            }
            const instructions = [];
            const signers = [];
            let shouldCreateMint = false;
            if (!targetMint) {
                signers.push(targetMintKeypair);
                targetMint = targetMintKeypair.publicKey;
                shouldCreateMint = true;
            }
            // Find the proper bonding index to use that isn't taken.
            let indexToUse = index || 0;
            const getTokenBonding = () => {
                return this.tokenBondingKey(targetMint, indexToUse);
            };
            const getTokenBondingAccount = () => __awaiter(this, void 0, void 0, function* () {
                return this.provider.connection.getAccountInfo((yield getTokenBonding())[0]);
            });
            if (!index) {
                // Find an empty voucher account
                while (yield getTokenBondingAccount()) {
                    indexToUse++;
                }
            }
            else {
                indexToUse = index;
            }
            const [targetMintAuthority, targetMintAuthorityBumpSeed] = yield web3_js_1.PublicKey.findProgramAddress([Buffer.from("target-authority", "utf-8"), targetMint.toBuffer()], this.programId);
            if (shouldCreateMint) {
                instructions.push(...(yield (0, common_1.createMintInstructions)(provider, targetMintAuthority, targetMint, targetMintDecimals)));
            }
            const [tokenBonding, bumpSeed] = yield this.tokenBondingKey(targetMint, indexToUse);
            let baseStorageAuthority = null;
            const [baseStorageAuthorityRes, baseStorageAuthorityBumpSeedRes] = yield this.baseStorageAuthorityKey(tokenBonding);
            const baseStorageAuthorityBumpSeed = baseStorageAuthorityBumpSeedRes;
            // This is a buy/sell bonding curve. Create the program owned base storage account
            if (!baseStorage) {
                baseStorageAuthority = baseStorageAuthorityRes;
                const baseStorageKeypair = anchor.web3.Keypair.generate();
                signers.push(baseStorageKeypair);
                baseStorage = baseStorageKeypair.publicKey;
                instructions.push(web3_js_1.SystemProgram.createAccount({
                    fromPubkey: payer,
                    newAccountPubkey: baseStorage,
                    space: spl_token_1.AccountLayout.span,
                    programId: spl_token_1.TOKEN_PROGRAM_ID,
                    lamports: yield this.provider.connection.getMinimumBalanceForRentExemption(spl_token_1.AccountLayout.span),
                }), spl_token_1.Token.createInitAccountInstruction(spl_token_1.TOKEN_PROGRAM_ID, baseMint, baseStorage, baseStorageAuthority));
            }
            let createdAccts = new Set();
            if (!buyTargetRoyalties) {
                buyTargetRoyalties = yield spl_token_1.Token.getAssociatedTokenAddress(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, targetMint, buyTargetRoyaltiesOwner, true);
                // If sell target royalties are undefined, we'll do this in the next step
                if (!createdAccts.has(buyTargetRoyalties.toBase58()) &&
                    !(yield this.accountExists(buyTargetRoyalties))) {
                    console.log("Creating buy target royalties...");
                    instructions.push(spl_token_1.Token.createAssociatedTokenAccountInstruction(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, targetMint, buyTargetRoyalties, buyTargetRoyaltiesOwner, payer));
                    createdAccts.add(buyTargetRoyalties.toBase58());
                }
            }
            if (!sellTargetRoyalties) {
                sellTargetRoyalties = yield spl_token_1.Token.getAssociatedTokenAddress(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, targetMint, sellTargetRoyaltiesOwner, true);
                if (!createdAccts.has(sellTargetRoyalties.toBase58()) &&
                    !(yield this.accountExists(sellTargetRoyalties))) {
                    console.log("Creating sell target royalties...");
                    instructions.push(spl_token_1.Token.createAssociatedTokenAccountInstruction(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, targetMint, sellTargetRoyalties, sellTargetRoyaltiesOwner, payer));
                    createdAccts.add(buyTargetRoyalties.toBase58());
                }
            }
            if (!buyBaseRoyalties) {
                buyBaseRoyalties = yield spl_token_1.Token.getAssociatedTokenAddress(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, baseMint, buyBaseRoyaltiesOwner, true);
                // If sell base royalties are undefined, we'll do this in the next step
                if (!createdAccts.has(buyBaseRoyalties.toBase58()) &&
                    !(yield this.accountExists(buyBaseRoyalties))) {
                    console.log("Creating base royalties...");
                    instructions.push(spl_token_1.Token.createAssociatedTokenAccountInstruction(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, baseMint, buyBaseRoyalties, buyBaseRoyaltiesOwner, payer));
                    createdAccts.add(buyBaseRoyalties.toBase58());
                }
            }
            if (!sellBaseRoyalties) {
                sellBaseRoyalties = yield spl_token_1.Token.getAssociatedTokenAddress(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, baseMint, sellBaseRoyaltiesOwner, true);
                if (!createdAccts.has(sellBaseRoyalties.toBase58()) &&
                    !(yield this.accountExists(sellBaseRoyalties))) {
                    console.log("Creating base royalties...");
                    instructions.push(spl_token_1.Token.createAssociatedTokenAccountInstruction(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, baseMint, sellBaseRoyalties, sellBaseRoyaltiesOwner, payer));
                    createdAccts.add(sellBaseRoyalties.toBase58());
                }
            }
            instructions.push(yield this.instruction.initializeTokenBondingV0({
                index: indexToUse,
                goLiveUnixTime: new bn_js_1.default(Math.floor(goLiveDate.valueOf() / 1000)),
                freezeBuyUnixTime: freezeBuyDate
                    ? new bn_js_1.default(Math.floor(freezeBuyDate.valueOf() / 1000))
                    : null,
                buyBaseRoyaltyPercentage: (0, spl_utils_1.percent)(buyBaseRoyaltyPercentage) || 0,
                buyTargetRoyaltyPercentage: (0, spl_utils_1.percent)(buyTargetRoyaltyPercentage) || 0,
                sellBaseRoyaltyPercentage: (0, spl_utils_1.percent)(sellBaseRoyaltyPercentage) || 0,
                sellTargetRoyaltyPercentage: (0, spl_utils_1.percent)(sellTargetRoyaltyPercentage) || 0,
                mintCap: mintCap || null,
                purchaseCap: purchaseCap || null,
                generalAuthority,
                curveAuthority,
                reserveAuthority,
                bumpSeed,
                baseStorageAuthority,
                baseStorageAuthorityBumpSeed,
                targetMintAuthorityBumpSeed,
                buyFrozen,
            }, {
                accounts: {
                    payer: payer,
                    curve,
                    tokenBonding,
                    baseMint,
                    targetMint: targetMint,
                    baseStorage,
                    buyBaseRoyalties,
                    buyTargetRoyalties,
                    sellBaseRoyalties,
                    sellTargetRoyalties,
                    tokenProgram: spl_token_1.TOKEN_PROGRAM_ID,
                    systemProgram: web3_js_1.SystemProgram.programId,
                    rent: web3_js_1.SYSVAR_RENT_PUBKEY,
                    clock: web3_js_1.SYSVAR_CLOCK_PUBKEY,
                },
            }));
            return {
                output: {
                    tokenBonding,
                    targetMint,
                    buyBaseRoyalties,
                    buyTargetRoyalties,
                    sellBaseRoyalties,
                    sellTargetRoyalties,
                    baseStorage,
                },
                instructions,
                signers,
            };
        });
    }
    /**
     * General utility function to check if an account exists
     * @param account
     * @returns
     */
    accountExists(account) {
        return __awaiter(this, void 0, void 0, function* () {
            return Boolean(yield this.provider.connection.getAccountInfo(account));
        });
    }
    /**
     * Runs {@link `createTokenBondingInstructions`}
     *
     * @param args
     * @returns
     */
    createTokenBonding(args) {
        return this.execute(this.createTokenBondingInstructions(args), args.payer);
    }
    /**
     * Update a bonding curve.
     *
     * @param param0
     * @returns
     */
    updateTokenBondingInstructions({ tokenBonding, buyBaseRoyaltyPercentage, buyTargetRoyaltyPercentage, sellBaseRoyaltyPercentage, sellTargetRoyaltyPercentage, buyBaseRoyalties, buyTargetRoyalties, sellBaseRoyalties, sellTargetRoyalties, generalAuthority, buyFrozen, }) {
        return __awaiter(this, void 0, void 0, function* () {
            const tokenBondingAcct = yield this.account.tokenBondingV0.fetch(tokenBonding);
            if (!tokenBondingAcct.generalAuthority) {
                throw new Error("Cannot update a token bonding account that has no authority");
            }
            const args = {
                buyBaseRoyaltyPercentage: (0, spl_utils_1.percent)(buyBaseRoyaltyPercentage) ||
                    tokenBondingAcct.buyBaseRoyaltyPercentage,
                buyTargetRoyaltyPercentage: (0, spl_utils_1.percent)(buyTargetRoyaltyPercentage) ||
                    tokenBondingAcct.buyTargetRoyaltyPercentage,
                sellBaseRoyaltyPercentage: (0, spl_utils_1.percent)(sellBaseRoyaltyPercentage) ||
                    tokenBondingAcct.sellBaseRoyaltyPercentage,
                sellTargetRoyaltyPercentage: (0, spl_utils_1.percent)(sellTargetRoyaltyPercentage) ||
                    tokenBondingAcct.sellTargetRoyaltyPercentage,
                generalAuthority: generalAuthority === null
                    ? null
                    : generalAuthority ||
                        tokenBondingAcct.generalAuthority,
                buyFrozen: typeof buyFrozen === "undefined"
                    ? tokenBondingAcct.buyFrozen
                    : buyFrozen,
            };
            return {
                output: null,
                signers: [],
                instructions: [
                    yield this.instruction.updateTokenBondingV0(args, {
                        accounts: {
                            tokenBonding,
                            generalAuthority: tokenBondingAcct.generalAuthority,
                            baseMint: tokenBondingAcct.baseMint,
                            targetMint: tokenBondingAcct.targetMint,
                            buyTargetRoyalties: buyTargetRoyalties || tokenBondingAcct.buyTargetRoyalties,
                            buyBaseRoyalties: buyBaseRoyalties || tokenBondingAcct.buyBaseRoyalties,
                            sellTargetRoyalties: sellTargetRoyalties || tokenBondingAcct.sellTargetRoyalties,
                            sellBaseRoyalties: sellBaseRoyalties || tokenBondingAcct.sellBaseRoyalties,
                        },
                    }),
                ],
            };
        });
    }
    /**
     * Runs {@link updateTokenBonding}
     * @param args
     */
    updateTokenBonding(args) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.execute(this.updateTokenBondingInstructions(args));
        });
    }
    /**
     * Instructions to buy twSOL from normal SOL.
     *
     * We wrap SOL so that the bonding contract isn't soaking up a bunch o SOL and damaging the security of the network.
     * The plan is to create a DAO for Strata that will govern what happens with this SOL.
     *
     * @param param0
     * @returns
     */
    buyBondingWrappedSolInstructions({ payer = this.wallet.publicKey, destination, source = this.wallet.publicKey, amount, }) {
        return __awaiter(this, void 0, void 0, function* () {
            const state = (yield this.getState());
            const stateAddress = (yield web3_js_1.PublicKey.findProgramAddress([Buffer.from("state", "utf-8")], this.programId))[0];
            const mintAuthority = (yield web3_js_1.PublicKey.findProgramAddress([Buffer.from("wrapped-sol-authority", "utf-8")], this.programId))[0];
            const mint = yield (0, common_1.getMintInfo)(this.provider, state.wrappedSolMint);
            let usedAta = false;
            if (!destination) {
                destination = yield spl_token_1.Token.getAssociatedTokenAddress(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, state.wrappedSolMint, source);
                usedAta = true;
            }
            const instructions = [];
            if (usedAta && !(yield this.accountExists(destination))) {
                instructions.push(spl_token_1.Token.createAssociatedTokenAccountInstruction(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, state.wrappedSolMint, destination, source, payer));
            }
            instructions.push(yield this.instruction.buyWrappedSolV0({
                amount: toBN(amount, mint),
            }, {
                accounts: {
                    state: stateAddress,
                    wrappedSolMint: state.wrappedSolMint,
                    mintAuthority: mintAuthority,
                    solStorage: state.solStorage,
                    source,
                    destination,
                    tokenProgram: spl_token_1.TOKEN_PROGRAM_ID,
                    systemProgram: web3_js_1.SystemProgram.programId,
                },
            }));
            return {
                signers: [],
                output: {
                    destination,
                },
                instructions,
            };
        });
    }
    /**
     * Invoke `buyBondingWrappedSol` instructions
     * @param args
     * @returns
     */
    buyBondingWrappedSol(args) {
        return this.execute(this.buyBondingWrappedSolInstructions(args), args.payer);
    }
    /**
     * Instructions to sell twSOL back into normal SOL.
     *
     * @param param0
     * @returns
     */
    sellBondingWrappedSolInstructions({ source, owner = this.wallet.publicKey, destination = this.wallet.publicKey, amount, all = false, }) {
        return __awaiter(this, void 0, void 0, function* () {
            const state = (yield this.getState());
            const stateAddress = (yield web3_js_1.PublicKey.findProgramAddress([Buffer.from("state", "utf-8")], this.programId))[0];
            const mint = yield (0, common_1.getMintInfo)(this.provider, state.wrappedSolMint);
            if (!source) {
                source = yield spl_token_1.Token.getAssociatedTokenAddress(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, state.wrappedSolMint, owner);
            }
            const instructions = [];
            instructions.push(yield this.instruction.sellWrappedSolV0({
                amount: toBN(amount, mint),
                all,
            }, {
                accounts: {
                    state: stateAddress,
                    wrappedSolMint: state.wrappedSolMint,
                    solStorage: state.solStorage,
                    source,
                    owner,
                    destination,
                    tokenProgram: spl_token_1.TOKEN_PROGRAM_ID,
                    systemProgram: web3_js_1.SystemProgram.programId,
                },
            }));
            if (all) {
                instructions.push(spl_token_1.Token.createCloseAccountInstruction(spl_token_1.TOKEN_PROGRAM_ID, source, destination, owner, []));
            }
            return {
                signers: [],
                output: null,
                instructions,
            };
        });
    }
    /**
     * Execute `sellBondingWrappedSolInstructions`
     * @param args
     * @returns
     */
    sellBondingWrappedSol(args) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.execute(this.sellBondingWrappedSolInstructions(args), args.payer);
        });
    }
    /**
     * Create a temporary account with `amount` twSOL, the token bonding wrapped sol mint.
     *
     * @param param0
     * @returns
     */
    createTemporaryWSolAccount({ payer, owner, amount, }) {
        return __awaiter(this, void 0, void 0, function* () {
            const state = (yield this.getState());
            const mint = yield (0, common_1.getMintInfo)(this.provider, state.wrappedSolMint);
            // Create a new account
            const newAccount = anchor.web3.Keypair.generate();
            const balanceNeeded = yield spl_token_1.Token.getMinBalanceRentForExemptAccount(this.provider.connection);
            return {
                firstInstructions: [
                    web3_js_1.SystemProgram.createAccount({
                        fromPubkey: payer,
                        newAccountPubkey: newAccount.publicKey,
                        lamports: balanceNeeded,
                        space: spl_token_1.AccountLayout.span,
                        programId: spl_token_1.TOKEN_PROGRAM_ID,
                    }),
                    spl_token_1.Token.createInitAccountInstruction(spl_token_1.TOKEN_PROGRAM_ID, state.wrappedSolMint, newAccount.publicKey, owner),
                    ...(yield this.buyBondingWrappedSolInstructions({
                        destination: newAccount.publicKey,
                        amount: toBN(amount, mint).add(new bn_js_1.default(1)),
                        source: owner,
                    })).instructions,
                ],
                lastInstructions: (yield this.sellBondingWrappedSolInstructions({
                    source: newAccount.publicKey,
                    owner,
                    all: true,
                    amount: toBN(amount, mint),
                })).instructions,
                signer: newAccount,
            };
        });
    }
    /**
     * Issue a command to buy `targetMint` tokens with `baseMint` tokens.
     *
     * @param param0
     * @returns
     */
    buyInstructions({ tokenBonding, source, sourceAuthority = this.wallet.publicKey, destination, desiredTargetAmount, baseAmount, slippage, payer = this.wallet.publicKey, }) {
        return __awaiter(this, void 0, void 0, function* () {
            const state = (yield this.getState());
            const tokenBondingAcct = yield this.account.tokenBondingV0.fetch(tokenBonding);
            // @ts-ignore
            const targetMint = yield (0, common_1.getMintInfo)(this.provider, tokenBondingAcct.targetMint);
            const baseMint = yield (0, common_1.getMintInfo)(this.provider, tokenBondingAcct.baseMint);
            const baseStorage = yield (0, common_1.getTokenAccount)(this.provider, tokenBondingAcct.baseStorage);
            // @ts-ignore
            const curve = yield this.getCurve(tokenBondingAcct.curve, baseStorage, baseMint, targetMint);
            const targetMintAuthority = yield web3_js_1.PublicKey.createProgramAddress([
                Buffer.from("target-authority", "utf-8"),
                tokenBondingAcct.targetMint.toBuffer(),
                new bn_js_1.default(tokenBondingAcct.targetMintAuthorityBumpSeed).toBuffer(),
            ], this.programId);
            const instructions = [];
            const signers = [];
            if (!destination) {
                destination = yield spl_token_1.Token.getAssociatedTokenAddress(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, tokenBondingAcct.targetMint, sourceAuthority);
                if (!(yield this.accountExists(destination))) {
                    console.log("Creating target account");
                    instructions.push(spl_token_1.Token.createAssociatedTokenAccountInstruction(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, tokenBondingAcct.targetMint, destination, sourceAuthority, payer));
                }
            }
            let buyTargetAmount = null;
            let buyWithBase = null;
            let rootEstimates = null;
            let maxPrice = 0;
            if (desiredTargetAmount) {
                const desiredTargetAmountNum = toNumber(desiredTargetAmount, targetMint);
                const neededAmount = desiredTargetAmountNum *
                    (1 / (1 - (0, curves_1.asDecimal)(tokenBondingAcct.buyTargetRoyaltyPercentage)));
                const curveAmount = curve.buyTargetAmount(desiredTargetAmountNum, tokenBondingAcct.buyBaseRoyaltyPercentage, tokenBondingAcct.buyTargetRoyaltyPercentage);
                maxPrice = curveAmount * (1 + slippage);
                rootEstimates = curve.buyTargetAmountRootEstimates(desiredTargetAmountNum, tokenBondingAcct.buyTargetRoyaltyPercentage);
                buyTargetAmount = {
                    targetAmount: new bn_js_1.default(Math.floor(neededAmount * Math.pow(10, targetMint.decimals))),
                    maximumPrice: toBN(maxPrice, baseMint),
                };
            }
            if (baseAmount) {
                const baseAmountNum = toNumber(baseAmount, baseMint);
                const min = curve.buyWithBaseAmount(baseAmountNum, tokenBondingAcct.buyBaseRoyaltyPercentage, tokenBondingAcct.buyTargetRoyaltyPercentage) *
                    (1 - slippage);
                maxPrice = baseAmountNum;
                rootEstimates = curve.buyWithBaseRootEstimates(baseAmountNum, tokenBondingAcct.buyBaseRoyaltyPercentage);
                buyWithBase = {
                    baseAmount: toBN(baseAmount, baseMint),
                    minimumTargetAmount: new bn_js_1.default(Math.ceil(min * Math.pow(10, targetMint.decimals))),
                };
            }
            let lastInstructions = [];
            if (!source) {
                if (tokenBondingAcct.baseMint.equals(state.wrappedSolMint)) {
                    const { signer, firstInstructions, lastInstructions: lastInstrs, } = yield this.createTemporaryWSolAccount({
                        payer: payer,
                        owner: sourceAuthority,
                        amount: maxPrice,
                    });
                    source = signer.publicKey;
                    signers.push(signer);
                    instructions.push(...firstInstructions);
                    lastInstructions.push(...lastInstrs);
                }
                else {
                    source = yield spl_token_1.Token.getAssociatedTokenAddress(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, tokenBondingAcct.baseMint, sourceAuthority);
                    if (!(yield this.accountExists(source))) {
                        throw new Error("Source account does not exist");
                    }
                }
            }
            const args = {
                // @ts-ignore
                buyTargetAmount,
                // @ts-ignore
                buyWithBase,
                rootEstimates: rootEstimates === null || rootEstimates === void 0 ? void 0 : rootEstimates.map(toU128),
            };
            const accounts = {
                accounts: {
                    tokenBonding,
                    // @ts-ignore
                    curve: tokenBondingAcct.curve,
                    baseMint: tokenBondingAcct.baseMint,
                    targetMint: tokenBondingAcct.targetMint,
                    targetMintAuthority,
                    baseStorage: tokenBondingAcct.baseStorage,
                    buyBaseRoyalties: tokenBondingAcct.buyBaseRoyalties,
                    buyTargetRoyalties: tokenBondingAcct.buyTargetRoyalties,
                    source,
                    sourceAuthority,
                    destination,
                    tokenProgram: spl_token_1.TOKEN_PROGRAM_ID,
                    clock: web3_js_1.SYSVAR_CLOCK_PUBKEY,
                },
            };
            instructions.push(yield this.instruction.buyV0(args, accounts));
            instructions.push(...lastInstructions);
            return {
                output: null,
                signers,
                instructions,
            };
        });
    }
    /**
     * Runs {@link buy}
     * @param args
     */
    buy(args) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.execute(this.buyInstructions(args), args.payer);
        });
    }
    getState() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.state) {
                return this.state;
            }
            const stateAddress = (yield web3_js_1.PublicKey.findProgramAddress([Buffer.from("state", "utf-8")], this.programId))[0];
            return this.account.programStateV0.fetchNullable(stateAddress);
        });
    }
    /**
     * Instructions to burn `targetMint` tokens in exchange for `baseMint` tokens
     *
     * @param param0
     * @returns
     */
    sellInstructions({ tokenBonding, source, sourceAuthority = this.wallet.publicKey, destination, targetAmount, slippage, payer = this.wallet.publicKey, }) {
        return __awaiter(this, void 0, void 0, function* () {
            const state = (yield this.getState());
            const tokenBondingAcct = yield this.account.tokenBondingV0.fetch(tokenBonding);
            if (tokenBondingAcct.sellFrozen) {
                throw new Error("Sell is frozen on this bonding curve");
            }
            // @ts-ignore
            const targetMint = yield (0, common_1.getMintInfo)(this.provider, tokenBondingAcct.targetMint);
            const baseMint = yield (0, common_1.getMintInfo)(this.provider, tokenBondingAcct.baseMint);
            const baseStorage = yield (0, common_1.getTokenAccount)(this.provider, tokenBondingAcct.baseStorage);
            // @ts-ignore
            const curve = yield this.getCurve(tokenBondingAcct.curve, baseStorage, baseMint, targetMint);
            let baseStorageAuthority;
            baseStorageAuthority = yield web3_js_1.PublicKey.createProgramAddress([
                Buffer.from("storage-authority", "utf-8"),
                tokenBonding.toBuffer(),
                new bn_js_1.default(tokenBondingAcct.baseStorageAuthorityBumpSeed).toBuffer(),
            ], this.programId);
            const instructions = [];
            const signers = [];
            if (!source) {
                source = yield spl_token_1.Token.getAssociatedTokenAddress(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, tokenBondingAcct.targetMint, sourceAuthority);
                if (!(yield this.accountExists(source))) {
                    throw new Error("Source account does not exist");
                }
            }
            const lastInstructions = [];
            if (!destination) {
                if (tokenBondingAcct.baseMint.equals(state.wrappedSolMint)) {
                    const { signer, firstInstructions, lastInstructions: lastInstrs, } = yield this.createTemporaryWSolAccount({
                        payer,
                        owner: sourceAuthority,
                        amount: 0,
                    });
                    destination = signer.publicKey;
                    signers.push(signer);
                    instructions.push(...firstInstructions);
                    lastInstructions.push(...lastInstrs);
                }
                else {
                    destination = yield spl_token_1.Token.getAssociatedTokenAddress(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, tokenBondingAcct.baseMint, sourceAuthority);
                    if (!(yield this.accountExists(destination))) {
                        console.log("Creating base account");
                        instructions.push(spl_token_1.Token.createAssociatedTokenAccountInstruction(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, tokenBondingAcct.baseMint, destination, sourceAuthority, payer));
                    }
                }
            }
            const targetAmountNum = toNumber(targetAmount, targetMint);
            const reclaimedAmount = curve.sellTargetAmount(targetAmountNum, tokenBondingAcct.sellBaseRoyaltyPercentage, tokenBondingAcct.sellTargetRoyaltyPercentage);
            const minPrice = Math.ceil(reclaimedAmount * (1 - slippage) * Math.pow(10, baseMint.decimals));
            const args = {
                targetAmount: toBN(targetAmount, targetMint),
                minimumPrice: new bn_js_1.default(minPrice),
                rootEstimates: curve
                    .buyTargetAmountRootEstimates(-targetAmountNum *
                    (1 - (0, curves_1.asDecimal)(tokenBondingAcct.sellTargetRoyaltyPercentage)), 0)
                    .map(toU128),
            };
            const accounts = {
                accounts: {
                    tokenBonding,
                    // @ts-ignore
                    curve: tokenBondingAcct.curve,
                    baseMint: tokenBondingAcct.baseMint,
                    targetMint: tokenBondingAcct.targetMint,
                    baseStorage: tokenBondingAcct.baseStorage,
                    sellBaseRoyalties: tokenBondingAcct.sellBaseRoyalties,
                    sellTargetRoyalties: tokenBondingAcct.sellTargetRoyalties,
                    baseStorageAuthority,
                    source,
                    sourceAuthority,
                    destination,
                    tokenProgram: spl_token_1.TOKEN_PROGRAM_ID,
                    clock: web3_js_1.SYSVAR_CLOCK_PUBKEY,
                },
            };
            instructions.push(yield this.instruction.sellV0(args, accounts));
            instructions.push(...lastInstructions);
            return {
                output: null,
                signers,
                instructions,
            };
        });
    }
    /**
     * Runs {@link sell}
     * @param args
     */
    sell(args) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.execute(this.sellInstructions(args), args.payer);
        });
    }
    /**
     * Get a class capable of displaying pricing information or this token bonding at its current reserve and supply
     *
     * @param tokenBonding
     * @returns
     */
    getPricing(tokenBonding) {
        return __awaiter(this, void 0, void 0, function* () {
            const tokenBondingAcct = yield this.account.tokenBondingV0.fetch(tokenBonding);
            const targetMint = yield (0, common_1.getMintInfo)(this.provider, tokenBondingAcct.targetMint);
            const baseMint = yield (0, common_1.getMintInfo)(this.provider, tokenBondingAcct.baseMint);
            const baseStorage = yield (0, common_1.getTokenAccount)(this.provider, tokenBondingAcct.baseStorage);
            return this.getCurve(tokenBondingAcct.curve, baseStorage, baseMint, targetMint);
        });
    }
    /**
     * Given some reserves and supply, get a pricing model for a curve at `key`.
     *
     * @param key
     * @param baseStorage
     * @param baseMint
     * @param targetMint
     * @returns
     */
    getCurve(key, baseStorage, baseMint, targetMint) {
        return __awaiter(this, void 0, void 0, function* () {
            const curve = yield this.account.curveV0.fetch(key);
            // @ts-ignore
            return (0, curves_1.fromCurve)(curve, baseStorage, baseMint, targetMint);
        });
    }
}
exports.SplTokenBonding = SplTokenBonding;
SplTokenBonding.ID = new web3_js_1.PublicKey("TBondz6ZwSM5fs4v2GpnVBMuwoncPkFLFR9S422ghhN");
SplTokenBonding.WRAPPED_SOL_MINT = new web3_js_1.PublicKey("Gjzxxh8UbFyrA4meWLSXGEA8UXeoMPZaFkCxn3WCo7V4");
//# sourceMappingURL=index.js.map